/*
    This is a romfs file system driver for Windows NT and Windows 2000.
    Copyright (C) 1999, 2000, 2001 Bo Brantn.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ntifs.h"
#include "fsd.h"
#include "rom_fs.h"
#include "border.h"

NTSTATUS
FsdDirectoryControl (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    NTSTATUS Status;

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

    switch (IrpContext->MinorFunction)
    {
    case IRP_MN_QUERY_DIRECTORY:
        Status = FsdQueryDirectory(IrpContext);
        break;

    default:
        Status = STATUS_INVALID_DEVICE_REQUEST;
        IrpContext->Irp->IoStatus.Status = Status;
        FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
        FsdFreeIrpContext(IrpContext);
    }

    return Status;
}

NTSTATUS
FsdQueryDirectory (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
    PDEVICE_OBJECT          DeviceObject;
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    PFSD_VCB                Vcb;
    PFILE_OBJECT            FileObject;
    PFSD_FCB                Fcb;
    PFSD_CCB                Ccb;
    PIRP                    Irp;
    PIO_STACK_LOCATION      IoStackLocation;
    FILE_INFORMATION_CLASS  FileInformationClass;
    ULONG                   Length;
    PUNICODE_STRING         FileName;
    ULONG                   FileIndex;
    BOOLEAN                 RestartScan;
    BOOLEAN                 ReturnSingleEntry;
    BOOLEAN                 IndexSpecified;
    PUCHAR                  Buffer;
    BOOLEAN                 FirstQuery;
    struct romfs_inode*     Inode = NULL;
    BOOLEAN                 FcbResourceAcquired = FALSE;
    ULONG                   QueryBlockLength;
    ULONG                   UsedLength = 0;
    USHORT                  InodeFileNameLength;
    UNICODE_STRING          InodeFileName;
    PULONG                  NextEntryOffset = NULL;

    InodeFileName.Buffer = NULL;

    __try
    {
        ASSERT(IrpContext);

        ASSERT((IrpContext->Identifier.Type == ICX) &&
               (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));

        DeviceObject = IrpContext->DeviceObject;

        //
        // This request is not allowed on the main device object
        //
        if (DeviceObject == FsdGlobalData.DeviceObject)
        {
            Status = STATUS_INVALID_DEVICE_REQUEST;
            __leave;
        }

        Vcb = (PFSD_VCB) DeviceObject->DeviceExtension;

        ASSERT(Vcb != NULL);

        ASSERT((Vcb->Identifier.Type == VCB) &&
               (Vcb->Identifier.Size == sizeof(FSD_VCB)));

        FileObject = IrpContext->FileObject;

        Fcb = (PFSD_FCB) FileObject->FsContext;

        ASSERT(Fcb);

        //
        // This request is not allowed on volumes
        //
        if (Fcb->Identifier.Type == VCB)
        {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        ASSERT((Fcb->Identifier.Type == FCB) &&
               (Fcb->Identifier.Size == sizeof(FSD_FCB)));

        if (!ExAcquireResourceSharedLite(
                 &Fcb->MainResource,
                 IrpContext->IsSynchronous
                 ))
        {
            Status = STATUS_PENDING;
            __leave;
        }

        FcbResourceAcquired = TRUE;

        //
        // This request is only allowed on directories
        //
        if (!(Fcb->FileAttributes & FILE_ATTRIBUTE_DIRECTORY))
        {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        Ccb = (PFSD_CCB) FileObject->FsContext2;

        ASSERT(Ccb);

        ASSERT((Ccb->Identifier.Type == CCB) &&
               (Ccb->Identifier.Size == sizeof(FSD_CCB)));

        Irp = IrpContext->Irp;

        IoStackLocation = IoGetCurrentIrpStackLocation(Irp);

#ifndef _GNU_NTIFS_

        FileInformationClass =
            IoStackLocation->Parameters.QueryDirectory.FileInformationClass;

        Length = IoStackLocation->Parameters.QueryDirectory.Length;

        FileName = IoStackLocation->Parameters.QueryDirectory.FileName;

        FileIndex = IoStackLocation->Parameters.QueryDirectory.FileIndex;

#else // _GNU_NTIFS_

        FileInformationClass = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileInformationClass;

        Length = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.Length;

        FileName = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileName;

        FileIndex = ((PEXTENDED_IO_STACK_LOCATION)
            IoStackLocation)->Parameters.QueryDirectory.FileIndex;

#endif // _GNU_NTIFS_

        RestartScan = IoStackLocation->Flags & SL_RESTART_SCAN;
        ReturnSingleEntry = IoStackLocation->Flags & SL_RETURN_SINGLE_ENTRY;
        IndexSpecified = IoStackLocation->Flags & SL_INDEX_SPECIFIED;

        if (!IrpContext->IsSynchronous)
        {
            Status = STATUS_PENDING;
            __leave;
        }

        Buffer = FsdGetUserBuffer(Irp);

        if (Buffer == NULL)
        {
            Status = STATUS_INVALID_USER_BUFFER;
            __leave;
        }

        if (FileName != NULL)
        {
            if (Ccb->DirectorySearchPattern.Buffer != NULL)
            {
                FirstQuery = FALSE;
            }
            else
            {
                FirstQuery = TRUE;

                Ccb->DirectorySearchPattern.Length =
                Ccb->DirectorySearchPattern.MaximumLength =
                    FileName->Length;

                Ccb->DirectorySearchPattern.Buffer =
                    FsdAllocatePool(NonPagedPool, FileName->Length);

                if (Ccb->DirectorySearchPattern.Buffer == NULL)
                {
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }

                RtlCopyMemory(
                    Ccb->DirectorySearchPattern.Buffer,
                    FileName->Buffer,
                    FileName->Length
                    );
            }
        }
        else if (Ccb->DirectorySearchPattern.Buffer != NULL)
        {
            FirstQuery = FALSE;
            FileName = &Ccb->DirectorySearchPattern;
        }
        else
        {
            FirstQuery = TRUE;

            Ccb->DirectorySearchPattern.Length =
            Ccb->DirectorySearchPattern.MaximumLength =
                sizeof(L"*");

            Ccb->DirectorySearchPattern.Buffer =
                FsdAllocatePool(NonPagedPool, sizeof(L"*"));

            if (Ccb->DirectorySearchPattern.Buffer == NULL)
            {
                Status = STATUS_INSUFFICIENT_RESOURCES;
                __leave;
            }

            RtlCopyMemory(
                Ccb->DirectorySearchPattern.Buffer,
                L"*",
                sizeof(L"*")
                );
        }

        if (!IndexSpecified)
        {
            if (RestartScan || FirstQuery)
            {
                FileIndex = Fcb->romfs_inode->spec;
            }
            else
            {
                FileIndex = Ccb->CurrentByteOffset;
            }
        }

        if (!(be32_to_cpu(FileIndex) & ROMFH_MASK))
        {
            if (FirstQuery)
            {
                Status = STATUS_NO_SUCH_FILE;
            }
            else
            {
                Status = STATUS_NO_MORE_FILES;
            }
            __leave;
        }

        Inode = (struct romfs_inode*) FsdAllocatePool(
            NonPagedPool,
            sizeof(struct romfs_inode) + ROMFS_MAXFN
            );

        if (Inode == NULL)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        RtlZeroMemory(Buffer, Length);

        switch (FileInformationClass)
        {
        case FileDirectoryInformation:
            if (Length < sizeof(FILE_DIRECTORY_INFORMATION))
            {
                Status = STATUS_INFO_LENGTH_MISMATCH;
                __leave;
            }
            QueryBlockLength = sizeof(FILE_DIRECTORY_INFORMATION);
            break;

        case FileFullDirectoryInformation:
            if (Length < sizeof(FILE_FULL_DIR_INFORMATION))
            {
                Status = STATUS_INFO_LENGTH_MISMATCH;
                __leave;
            }
            QueryBlockLength = sizeof(FILE_FULL_DIR_INFORMATION);
            break;

        case FileBothDirectoryInformation:
            if (Length < sizeof(FILE_BOTH_DIR_INFORMATION))
            {
                Status = STATUS_INFO_LENGTH_MISMATCH;
                __leave;
            }
            QueryBlockLength = sizeof(FILE_BOTH_DIR_INFORMATION);
            break;

        case FileNamesInformation:
            if (Length < sizeof(FILE_NAMES_INFORMATION))
            {
                Status = STATUS_INFO_LENGTH_MISMATCH;
                __leave;
            }
            QueryBlockLength = sizeof(FILE_NAMES_INFORMATION);
            break;

        default:
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }

        while (UsedLength < Length
           && (be32_to_cpu(FileIndex) & ROMFH_MASK))
        {
            Status = FsdReadInodeByIndex(
                Vcb->TargetDeviceObject,
                FileIndex,
                Inode
                );

            if (!NT_SUCCESS(Status))
            {
                if (!UsedLength)
                {
                    if (FirstQuery)
                    {
                        Status = STATUS_NO_SUCH_FILE;
                    }
                    else
                    {
                        Status = STATUS_NO_MORE_FILES;
                    }
                }
                else
                {
                    Status = STATUS_SUCCESS;
                }
                __leave;
            }

            InodeFileNameLength = strnlen(Inode->name, ROMFS_MAXFN);

            if (Length - UsedLength < QueryBlockLength +
                InodeFileNameLength * 2 - sizeof(WCHAR))
            {
                if (!UsedLength)
                {
                    Status = STATUS_INFO_LENGTH_MISMATCH;
                }
                else
                {
                    Status = STATUS_SUCCESS;
                }
                __leave;
            }

            InodeFileName.Length =
            InodeFileName.MaximumLength =
                InodeFileNameLength * 2;

            InodeFileName.Buffer = FsdAllocatePool(
                NonPagedPool,
                InodeFileNameLength * 2
                );

            FsdCharToWchar(
                InodeFileName.Buffer,
                Inode->name,
                InodeFileNameLength
                );

            if (FsRtlIsNameInExpression(
                FileName,
                &InodeFileName,
                FALSE,
                NULL
                ))
            {
                switch (FileInformationClass)
                {
                case FileDirectoryInformation:
                    {
                        PFILE_DIRECTORY_INFORMATION
                            FileDirectoryInformation;

                        FileDirectoryInformation =
                            (PFILE_DIRECTORY_INFORMATION) (Buffer + UsedLength);

                        FileDirectoryInformation->FileIndex = FileIndex;

                        FileDirectoryInformation->CreationTime.QuadPart =
                            0;

                        FileDirectoryInformation->LastAccessTime.QuadPart =
                            0;

                        FileDirectoryInformation->LastWriteTime.QuadPart =
                            0;

                        FileDirectoryInformation->ChangeTime.QuadPart =
                            0;

                        FileDirectoryInformation->EndOfFile.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileDirectoryInformation->AllocationSize.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileDirectoryInformation->FileAttributes =
                            FILE_ATTRIBUTE_NORMAL;

                        if (be32_to_cpu(Inode->next) & ROMFH_DIR)
                        {
                            FileDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_DIRECTORY;
                        }

#ifndef FSD_RO
                        if (Vcb->ReadOnly)
#endif
                        {
                            FileDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_READONLY;
                        }

                        FileDirectoryInformation->FileNameLength =
                            InodeFileNameLength * 2;

                        RtlCopyMemory(
                            FileDirectoryInformation->FileName,
                            InodeFileName.Buffer,
                            InodeFileNameLength * 2
                            );

                        UsedLength += QueryBlockLength +
                            InodeFileNameLength * 2 - sizeof(WCHAR);

                        if (!ReturnSingleEntry)
                        {
                            FileDirectoryInformation->NextEntryOffset =
                                QueryBlockLength + InodeFileNameLength * 2 -
                                sizeof(WCHAR) + UsedLength % 8;
                        }
                        else
                        {
                            FileDirectoryInformation->NextEntryOffset = 0;
                        }

                        UsedLength += UsedLength % 8;

                        NextEntryOffset =
                            &FileDirectoryInformation->NextEntryOffset;
                    }
                    break;

                case FileFullDirectoryInformation:
                    {
                        PFILE_FULL_DIR_INFORMATION
                            FileFullDirectoryInformation;

                        FileFullDirectoryInformation =
                            (PFILE_FULL_DIR_INFORMATION) (Buffer + UsedLength);

                        FileFullDirectoryInformation->FileIndex = FileIndex;

                        FileFullDirectoryInformation->CreationTime.QuadPart =
                            0;

                        FileFullDirectoryInformation->LastAccessTime.QuadPart =
                            0;

                        FileFullDirectoryInformation->LastWriteTime.QuadPart =
                            0;

                        FileFullDirectoryInformation->ChangeTime.QuadPart =
                            0;

                        FileFullDirectoryInformation->EndOfFile.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileFullDirectoryInformation->AllocationSize.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileFullDirectoryInformation->FileAttributes =
                            FILE_ATTRIBUTE_NORMAL;

                        if (be32_to_cpu(Inode->next) & ROMFH_DIR)
                        {
                            FileFullDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_DIRECTORY;
                        }

#ifndef FSD_RO
                        if (Vcb->ReadOnly)
#endif
                        {
                            FileFullDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_READONLY;
                        }

                        FileFullDirectoryInformation->FileNameLength =
                            InodeFileNameLength * 2;

                        FileFullDirectoryInformation->EaSize = 0;

                        RtlCopyMemory(
                            FileFullDirectoryInformation->FileName,
                            InodeFileName.Buffer,
                            InodeFileNameLength * 2
                            );

                        UsedLength += QueryBlockLength +
                            InodeFileNameLength * 2 - sizeof(WCHAR);

                        if (!ReturnSingleEntry)
                        {
                            FileFullDirectoryInformation->NextEntryOffset =
                                QueryBlockLength + InodeFileNameLength * 2 -
                                sizeof(WCHAR) + UsedLength % 8;
                        }
                        else
                        {
                            FileFullDirectoryInformation->NextEntryOffset = 0;
                        }

                        UsedLength += UsedLength % 8;

                        NextEntryOffset =
                            &FileFullDirectoryInformation->NextEntryOffset;
                    }
                    break;

                case FileBothDirectoryInformation:
                    {
                        PFILE_BOTH_DIR_INFORMATION
                            FileBothDirectoryInformation;

                        FileBothDirectoryInformation =
                            (PFILE_BOTH_DIR_INFORMATION) (Buffer + UsedLength);

                        FileBothDirectoryInformation->FileIndex = FileIndex;

                        FileBothDirectoryInformation->CreationTime.QuadPart =
                            0;

                        FileBothDirectoryInformation->LastAccessTime.QuadPart =
                            0;

                        FileBothDirectoryInformation->LastWriteTime.QuadPart =
                            0;

                        FileBothDirectoryInformation->ChangeTime.QuadPart =
                            0;

                        FileBothDirectoryInformation->EndOfFile.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileBothDirectoryInformation->AllocationSize.QuadPart =
                            be32_to_cpu(Inode->size);

                        FileBothDirectoryInformation->FileAttributes =
                            FILE_ATTRIBUTE_NORMAL;

                        if (be32_to_cpu(Inode->next) & ROMFH_DIR)
                        {
                            FileBothDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_DIRECTORY;
                        }

#ifndef FSD_RO
                        if (Vcb->ReadOnly)
#endif
                        {
                            FileBothDirectoryInformation->FileAttributes |=
                                FILE_ATTRIBUTE_READONLY;
                        }

                        FileBothDirectoryInformation->FileNameLength =
                            InodeFileNameLength * 2;

                        FileBothDirectoryInformation->EaSize = 0;

                        // TODO: Present a short alias

                        // Here we would like to use RtlGenerate8dot3Name but
                        // I don't know how to use the argument
                        // PGENERATE_NAME_CONTEXT

                        // FileBothDirectoryInformation->ShortNameLength
                        // FileBothDirectoryInformation->ShortName

                        RtlCopyMemory(
                            FileBothDirectoryInformation->FileName,
                            InodeFileName.Buffer,
                            InodeFileNameLength * 2
                            );

                        UsedLength += QueryBlockLength +
                            InodeFileNameLength * 2 - sizeof(WCHAR);

                        if (!ReturnSingleEntry)
                        {
                            FileBothDirectoryInformation->NextEntryOffset =
                                QueryBlockLength + InodeFileNameLength * 2 -
                                sizeof(WCHAR) + UsedLength % 8;
                        }
                        else
                        {
                            FileBothDirectoryInformation->NextEntryOffset = 0;
                        }

                        UsedLength += UsedLength % 8;

                        NextEntryOffset =
                            &FileBothDirectoryInformation->NextEntryOffset;
                    }
                    break;

                case FileNamesInformation:
                    {
                        PFILE_NAMES_INFORMATION
                            FileNamesInformation;

                        FileNamesInformation =
                            (PFILE_NAMES_INFORMATION) (Buffer + UsedLength);

                        FileNamesInformation->FileIndex = FileIndex;

                        FileNamesInformation->FileNameLength =
                            InodeFileNameLength * 2;

                        RtlCopyMemory(
                            FileNamesInformation->FileName,
                            InodeFileName.Buffer,
                            InodeFileNameLength * 2
                            );

                        UsedLength += QueryBlockLength +
                            InodeFileNameLength * 2 - sizeof(WCHAR);

                        if (!ReturnSingleEntry)
                        {
                            FileNamesInformation->NextEntryOffset =
                                QueryBlockLength + InodeFileNameLength * 2 -
                                sizeof(WCHAR) + UsedLength % 8;
                        }
                        else
                        {
                            FileNamesInformation->NextEntryOffset = 0;
                        }

                        UsedLength += UsedLength % 8;

                        NextEntryOffset =
                            &FileNamesInformation->NextEntryOffset;
                    }
                    break;
                }
            }

            if (InodeFileName.Buffer != NULL)
            {
                FsdFreePool(InodeFileName.Buffer);
                InodeFileName.Buffer = NULL;
            }

            Ccb->CurrentByteOffset =
            FileIndex =
                Inode->next;

            if (UsedLength && ReturnSingleEntry)
            {
                Status = STATUS_SUCCESS;
                __leave;
            }
        }

        if (!UsedLength)
        {
            if (FirstQuery)
            {
                Status = STATUS_NO_SUCH_FILE;
            }
            else
            {
                Status = STATUS_NO_MORE_FILES;
            }
        }
        else
        {
            Status = STATUS_SUCCESS;
        }
    }
    __finally
    {
        if (FcbResourceAcquired)
        {
            ExReleaseResourceForThreadLite(
                &Fcb->MainResource,
                ExGetCurrentResourceThread()
                );
        }

        if (Inode != NULL)
        {
            FsdFreePool(Inode);
        }

        if (InodeFileName.Buffer != NULL)
        {
            FsdFreePool(InodeFileName.Buffer);
        }

        if (NextEntryOffset != NULL)
        {
            *NextEntryOffset = 0;
        }

        if (!IrpContext->ExceptionInProgress)
        {
            if (Status == STATUS_PENDING)
            {
                Status = FsdLockUserBuffer(
                    IrpContext->Irp,
                    Length,
                    IoWriteAccess
                    );

                if (NT_SUCCESS(Status))
                {
                    Status = FsdQueueRequest(IrpContext);
                }
                else
                {
                    IrpContext->Irp->IoStatus.Status = Status;
                    FsdCompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
                    FsdFreeIrpContext(IrpContext);
                }
            }
            else
            {
                IrpContext->Irp->IoStatus.Information = UsedLength;
                IrpContext->Irp->IoStatus.Status = Status;

                FsdCompleteRequest(
                    IrpContext->Irp,
                    (CCHAR)
                    (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT)
                    );

                FsdFreeIrpContext(IrpContext);
            }
        }
    }

    return Status;
}
